import math
import networkx as nx
from matplotlib import pyplot as plt
import numpy as np
import torch
from torch import nn

import graph_walker

import fire


class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout=0.0, max_len=5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)


class Model(nn.Module):
    def __init__(self, num_layers, walk_len, dim=128):
        super().__init__()
        print(f'Using {num_layers} layers and {dim} dimensions')
        self.walk_len = walk_len
        self.dim = dim
        self.walk_embed = nn.Embedding(20, dim)
        self.attr_embed = nn.Linear(1, dim)
        self.pos_encoder = PositionalEncoding(d_model=dim, dropout=0.)
        assert num_layers == 1
        self.encoder = nn.Sequential(
            nn.Linear(walk_len * dim, dim),
            nn.ReLU(),
            nn.Linear(dim, dim),
            nn.ReLU(),
            nn.Linear(dim, dim),
            nn.ReLU()
        )
        self.head = nn.Linear(dim, 1)

    def forward(self, named_walks, attr):
        x = self.walk_embed(named_walks - 1) + self.attr_embed(attr[..., None])
        x = self.pos_encoder(x)
        x = self.encoder(x.view(x.shape[0], self.walk_len * self.dim))
        return self.head(x).squeeze(-1)


def get_data(N_A, N_B, device):
    x_A = torch.rand(N_A) * -math.sqrt(3)
    x_B = torch.rand(N_B) * math.sqrt(3)
    y_A = torch.empty(N_A).fill_(x_B.mean())
    y_B = torch.empty(N_B).fill_(x_A.mean())
    return torch.cat([x_A, x_B]).to(device), torch.cat([y_A, y_B]).to(device)


def main(
    experiment='barbell',
    N=10,
    min_degree=False,
    no_backtrack=False,
    walk_len=1000,
    train_n_walks=8,
    eval_n_walks=8,
    seed=0,
):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    if experiment == 'barbell':
        G = nx.to_undirected(nx.barbell_graph(N, 0))
        for i in range(2 * N):
            G.nodes[i]['label'] = 0 if i < N else 1
        nx.draw(G, with_labels=True, node_color=['r' if G.nodes[i]['label'] == 0 else 'b' for i in G.nodes])
        plt.savefig('experiments/barbell_graph.png')
    elif experiment == 'clique':
        G = nx.to_undirected(nx.complete_graph(2 * N))
        for i in range(2 * N):
            G.nodes[i]['label'] = 0 if i < N else 1
        nx.draw(G, with_labels=True, node_color=['r' if G.nodes[i]['label'] == 0 else 'b' for i in G.nodes])
        plt.savefig('experiments/clique_graph.png')
    else:
        raise ValueError(f'Unknown experiment {experiment}')

    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    train_dataset = [get_data(N, N, device) for _ in range(100)]
    test_dataset = [get_data(N, N, device) for _ in range(100)]

    model = Model(num_layers=1 if experiment == 'barbell' else 1, walk_len=walk_len).to(device)
    optimizer = torch.optim.AdamW(model.parameters(), lr=0.001, weight_decay=0.01)

    for epoch in range(500):
        model.train()
        train_losses = []
        for train_data_idx in torch.randperm(100):
            x, y = train_dataset[train_data_idx]
            walks, _ = graph_walker.random_walks(G=G, n_walks=train_n_walks, walk_len=walk_len, min_degree=min_degree, no_backtrack=no_backtrack, verbose=False)
            named_walks = torch.tensor(graph_walker._anonymize(walks).astype(np.int32), dtype=torch.long, device=device)
            walks = walks.astype(np.int32)
            attr = x[walks]
            targets = y[walks[:, 0]]

            optimizer.zero_grad()
            outputs = model.forward(named_walks, attr)
            loss = nn.functional.mse_loss(outputs, targets) * 4 * N
            loss.backward()
            optimizer.step()
            train_losses.append(loss.item())

        model.eval()
        test_losses = []
        base_1_losses = []
        base_2_losses = []
        for test_data in test_dataset:
            x, y = test_data
            walks, _ = graph_walker.random_walks(G=G, n_walks=eval_n_walks, walk_len=walk_len, min_degree=min_degree, no_backtrack=no_backtrack, verbose=False)
            named_walks = torch.tensor(graph_walker._anonymize(walks).astype(np.int32), dtype=torch.long, device=device)
            walks = walks.astype(np.int32)
            attr = x[walks]
            targets = y[walks[:, 0]]

            outputs = model.forward(named_walks, attr)
            # Monte Carlo averaging
            outputs = outputs.view(eval_n_walks, -1).mean(dim=0)
            targets = targets.view(eval_n_walks, -1).mean(dim=0)

            loss = nn.functional.mse_loss(outputs, targets) * 4 * N
            test_losses.append(loss.item())

            base_1_loss = nn.functional.mse_loss(torch.zeros_like(targets), targets) * 4 * N
            base_2_loss = nn.functional.mse_loss(targets.sign() * math.sqrt(3) / 2, targets) * 4 * N
            base_1_losses.append(base_1_loss.item())
            base_2_losses.append(base_2_loss.item())

        print(f'Epoch {epoch} | Train Loss {np.mean(train_losses):.4f} | Test Loss {np.mean(test_losses):.4f} | Base 1 Loss {np.mean(base_1_losses):.4f} | Base 2 Loss {np.mean(base_2_losses):.4f}')

    torch.save(model.state_dict(), f'experiments/mlp_experiment_{experiment}_N={N}_walk_len={walk_len}_train_n_walks={train_n_walks}_eval_n_walks={eval_n_walks}_seed={seed}.pth')


if __name__ == '__main__':
    fire.Fire(main)
